/*

ImageBackgroundMean.js: Image background mean calculator
================================================================================

Das Script ermittelt den Background besser als die StandardFormel
	bg = 3 * median – 2 * mean.
Es listet auf der Console den Wert als normalisierte und als 16-Bit Wert an.

The script determines the background better than the Standard formula
	bg = 3 * median - 2 * mean.
The values are listed as normalized and as a 16-bit value (Process Console).
================================================================================


 */
#define TITLE "ImageBackgroundMean"
#define VERSION "1.0"

#feature-id Image Analysis > ImageBackgroundMean

#feature-info An image background mean calculator.<br/>\
   <br/>\
   This calculation is based on the photometric library DAOPHOT\
   and the code was translated from sub MMM written in IDL.\
   <br/>\
   Copyright &copy; 2014 Hartmut V. Bornemann. All Rights Reserved.

function ImageBackgroundMean(img, channel, mxiter, readnoise)
{
   // This is a translation from DAOPHOT sub MMM
   //
   // INPUTS:
   //       img - Image with vector containing sky values. This version of
   //             MMM does not require SKY to be sorted beforehand.
   //             img is unaltered by this program.
   // OUTPUTS
   // skymod - Scalar giving estimated mode of the sky values
   // SIGMA -  Scalar giving standard deviation of the peak in the sky
   //          histogram.  If for some reason it is impossible to derive
   //          skymod, then SIGMA = -1.0
   // SKEW -   Scalar giving skewness of the peak in the sky histogram

   if (mxiter == 0) mxiter = 50;
   var minsky = 20;
   img.selectedChannel = channel;
   var sky = new Array(img.width * img.height);
   img.getPixels (sky);

   var nsky = sky.length;
   var nlast = nsky-1;        // Subscript of last pixel in SKY array
   sky.sort(function (a, b) {return a > b;}); // Sort SKY in ascending values
   // Median of all sky values
   var skymid = 0.5*sky[Int((nsky-1)/2)] + 0.5*sky[Int(nsky/2)];
   var cut = Math.min( skymid-sky[0],sky[nsky-1] - skymid );
   var cut2 = skymid + cut;
   var cut1 = skymid - cut;

   // Select the pixels between Cut1 and Cut2

   var j1 = -1;
   var j2 = 0;

   for (var i = 0; i < nsky; i++)
   {
      if (sky[i] >= cut1)
      {
         j1 = i;
         j2 = i;
         for (var j = 0; j < nsky; j++)
         {
            if (sky[j] <= cut2)
               j2 = j;
            else
               break;
         }
         break;
      }
   }

   if (j1 < 0)
   {
      Console.writeln("ERROR - No sky values fall within " +
                        cut1.toFixed(2) + " and " + cut2.toFixed(3));
      return -1;
   }

   var good = j2 - j1 + 1;

   var delta  = new Array(good);
   for (var i = 0; i < good; i++)
   {
      delta[i] = sky[j1 + i] - skymid; //Subtract median to improve arithmetic accuracy
   }
   var sum = total(delta);
   var sumsq = totalSqrd(delta);
   // Highest value accepted at upper end of vector
   var maximm = j2;
   // Highest value reject at lower end of vector
   var minimm = j1 - 1;
   // Compute mean and sigma (from the first pass).
   var skymed = 0.5*sky[Int((minimm+maximm+1)/2)] +
                0.5*sky[Int((minimm+maximm)/2 ) + 1];   // median
   var skymn = sum/(maximm-minimm);                     // mean
   var sigma = Math.sqrt(sumsq/(maximm-minimm)-Math.pow(skymn,2)); // sigma
   skymn = skymn + skymid; // Add median which was subtracted off earlier
   // If mean is less than the mode, then the contamination is slight, and the
   //    mean value is what we really want.
   var skymod =  (skymed < skymn) ? 3.0 * skymed - 2.0 * skymn : skymn;
   // Console.writeln("skymod "+skymod);
   // Rejection and recomputation loop:
   var clamp = 1;
   var old   = 0;
   var niter = 0;
   while (true)
   {
      niter += 1;
      if (niter > mxiter)
      {
         Console.writeln("ERROR - Too many (" + mxiter + ") iterations, " +
                         "unable to compute sky");
         return -1;
      }
      if ( (maximm-minimm) < minsky )
      {
         Console.writeln("ERROR - Too few ("+ (maximm-minimm).toString() +
                         ") valid sky elements, unable to compute sky");
         return -1;
      }
      // Compute Chauvenet rejection criterion
      var  r = Math.log10( maximm-minimm );
      r = Math.max( 2.0, ( -0.1042*r + 1.1695)*r + 0.8895 );
      // Compute rejection limits (symmetric about the current mode)
      var cut = r*sigma + 0.5*Math.abs(skymn-skymod);
      cut1 = skymod - cut;
      cut2 = skymod + cut;
      // Recompute mean and sigma by adding and/or subtracting sky values
      // at both ends of the interval of acceptable values.
      var redo = false;
      var newmin = minimm;
      var tst_min = sky[newmin+1] >= cut1; // Is minimm+1 above current CUT?
      var done = (newmin == -1) & tst_min;   // Are we at first pixel of SKY?
      if (!done) done = (sky[Math.max(newmin, 0)] < cut1) & tst_min;
      if (!done)
      {
         var istep = 1 - 2 * boolToInteger(tst_min);
         while (!done)
         {
            newmin = newmin + istep;
            done = (newmin == -1) || (newmin == nlast);
            if (!done) done = (sky[newmin] <= cut1) && (sky[newmin+1] >= cut1);
         }
         if (tst_min)
            delta = SkySub(sky, newmin + 1, minimm, skymid);
         else
            delta = SkySub(sky, minimm + 1, newmin, skymid);

         sum   -= istep * total(delta);
         sumsq -= istep * totalSqrd(delta);
         redo   = true;
         minimm = newmin;
      }

      //

      var newmax = maximm;
      var tst_max = sky[maximm] <= cut2;  // Is current maximum below upper cut?
      done = (maximm == nlast) && tst_max; // Are we at last pixel of SKY array?
      if (!done) done = tst_max && (sky[Math.min((maximm+1), nlast)] > cut2);
      if (!done)
      {
         istep = -1 + 2 * boolToInteger(tst_max); // Increment up or down?
         while (!done)
         {
            newmax = newmax + istep;
            done = (newmax == nlast) || (newmax == -1);
            if (!done)
               done = ( sky[newmax] <= cut2 ) && ( sky[newmax+1] >= cut2 );
         }
         if (tst_max)
            delta = SkySub(sky, maximm + 1, newmax, skymid);
         else
            delta = SkySub(sky, newmax + 1, maximm, skymid);
         sum   += istep * total(delta);
         sumsq += istep * totalSqrd(delta);
         redo   = true;
         maximm = newmax;
      }
     /*
         Compute mean and sigma (from this pass).
     */
      nsky = maximm - minimm;
      if ( nsky < minsky )
      {
         Console.writeln("ERROR - Outlier rejection left too few sky elements");
         return -1;
      }

      skymn = sum/nsky;
      sigma = Math.sqrt(Math.max(sumsq / nsky - Math.pow(skymn, 2), 0));
      skymn = skymn + skymid;
      /*
            Determine a more robust median by averaging the central 20% of pixels.
            Estimate the median using the mean of the central 20 percent of sky
            values.   Be careful to include a perfectly symmetric sample of pixels about
            the median, whether the total number is even or odd within the acceptance
            interval
      */
      var center = (minimm + 1 + maximm)/2.0;
      var side = Math.round(0.2 * (maximm - minimm)) / 2.0 + 0.25;
      var J = Math.round(center - side);
      var K = Math.round(center + side);
      /*
         In case  the data has a large number of of the same (quantized)
         intensity, expand the range until both limiting values differ from the
         central value by at least 0.25 times the read noise.
      */
      if (readnoise > 0)
      {
         var L = Math.round(center - 0.25);
         var M = Math.round(center + 0.25);
         while ((J > K) & (K < nsky - 1) &
                      (((sky[L] - sky[J]) < r) | ((sky[K] - sky[M]) < r)))
         {
            J -= 1
            K += 1
         }
      }

      skymed = total(SkySub(sky, J, K, 0)) / (K - J + 1);

      /*
         If the mean is less than the median, then the problem of contamination
         is slight, and the mean is what we really want.
      */

      var dmod;
      if (skymed < skymn)
         dmod = 3.0*skymed-2.0*skymn-skymod;
      else
         dmod = skymn - skymod;
      // prevent oscillations by clamping down if sky adjustments are changing sign
      if (dmod*old < 0) clamp = 0.5*clamp;
      skymod = skymod + clamp*dmod;
      old = dmod;

      if(!redo) break;
   }

   // for later use:
   // skew = (skymn - skymod) / Math.max(1.0, sigma);
   // nsky = maximm - minimm;

   return skymod;
}

function SkySub( Values, StartIndex, EndIndex, Reduction)
{
   //
   // copy a subarray and subtract a constant value
   //
   var v = new Array(EndIndex - StartIndex + 1);
   for (var i = 0; i <= v.length - 1; i++)
   {
      v[i] = Values[i + StartIndex] - Reduction;
   }
   return v;
}

function boolToInteger(b)
{
   if (b)
      return 1;
   else
      return 0;
}

function Int(f)
{
   if (f < 0)
      return Math.ceil(f);
   else
      return Math.floor(f);
}

function total(array)
{
   var sum = 0;
   for (var i=0; i<array.length; i++)
   {
      sum += array[i];
   }
   return sum;
}

function totalSqrd(array)
{
   var sum = 0;
   for (var i=0; i<array.length; i++)
   {
      sum += Math.pow(array[i], 2);
   }
   return sum;
}


function main()
{
   var window = ImageWindow.activeWindow;
   if ( window.isNull ) throw new Error( "No active image" );
   Console.show();
   Console.writeln("\r\nImage " + window.mainView.id);
   Console.writeln("ImageBackgroundMean calculation:");
   Console.flush();
   for (var channel = 0; channel < window.mainView.image.numberOfChannels; channel++)
   {
      var skymod = ImageBackgroundMean(window.mainView.image, channel, 0, 0.1e-004);
      if (skymod > 0)
      {
         Console.write("Channel " + channel);
         Console.write(", BackgroundMean (normalized) = " + skymod.toFixed(8));
         Console.writeln(", (16-Bit) = " + (skymod * 65535).toFixed(8));
         Console.flush();
      }
   }
 }

main();
